from typing import Any, Union, Literal, Optional, TypeAlias

# A lot of annotation features a depend on the Python version:
# - `typing.TypeAlias` Type aliases are supported from Python 3.10 to 3.11
# - `type` Type statements are supported from Python 3.12, replacing `typing.TypeAlias`
# - `typing.Literal` Literal types are supported from Python 3.8
#
# We can't and shouldn't use a different `.pyi` file for every single Python version.
# So let's assume we are targeting Python 3.11 and we have NumPy available.
from numpy.typing import NDArray

_BufferType: TypeAlias = Union[NDArray[Any], memoryview]

_MetricType = Literal[
    "euclidean",
    "sqeuclidean",
    "inner",
    "dot",
    "cosine",
    "cos",
    "hamming",
    "jaccard",
    "kullbackleibler",
    "kl",
    "jensenshannon",
    "js",
    "intersection",
    "bilinear",
    "mahalanobis",
    "fma",
    "wsum",
]
_IntegralType = Literal[
    # Booleans
    "c",
    "b8",
    # Signed integers
    "b",
    "i8",
    "int8",
    "h",
    "i16",
    "int16",
    "i",
    "l",
    "i32",
    "int32",
    "q",
    "i64",
    "int64",
    # Unsigned integers
    "B",
    "u8",
    "uint8",
    "H",
    "u16",
    "uint16",
    "I",
    "L",
    "u32",
    "uint32",
    "Q",
    "u64",
    "uint64",
]
_FloatType = Literal[
    "f",
    "f32",
    "float32",
    "e",
    "f16",
    "float16",
    "d",
    "f64",
    "float64",
    "bh",  #! Not supported by NumPy
    "bf16",  #! Not supported by NumPy
    "bfloat16",  #! Not supported by NumPy
]
_ComplexType = Literal[
    "complex32",  #! Not supported by NumPy
    "bcomplex32",  #! Not supported by NumPy
    "complex64",
    "complex128",
]

class DistancesTensor(memoryview): ...

# ---------------------------------------------------------------------

# Controlling SIMD capabilities
def get_capabilities() -> dict[str, bool]: ...
def enable_capability(capability: str, /) -> None: ...
def disable_capability(capability: str, /) -> None: ...

# Accessing function pointers
def pointer_to_euclidean(dtype: Union[_IntegralType, _FloatType], /) -> int: ...
def pointer_to_sqeuclidean(dtype: Union[_IntegralType, _FloatType], /) -> int: ...
def pointer_to_cosine(dtype: Union[_IntegralType, _FloatType], /) -> int: ...
def pointer_to_inner(dtype: Union[_FloatType, _ComplexType], /) -> int: ...
def pointer_to_dot(dtype: Union[_FloatType, _ComplexType], /) -> int: ...
def pointer_to_vdot(dtype: Union[_FloatType, _ComplexType], /) -> int: ...
def pointer_to_hamming(dtype: _IntegralType, /) -> int: ...
def pointer_to_jaccard(dtype: _IntegralType, /) -> int: ...
def pointer_to_jensenshannon(dtype: _FloatType, /) -> int: ...
def pointer_to_kullbackleibler(dtype: _FloatType, /) -> int: ...

# ---------------------------------------------------------------------

# All pairwise distances, similar to: `scipy.spatial.distance.cdist`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.cdist.html
def cdist(
    a: _BufferType,
    b: _BufferType,
    /,
    metric: _MetricType = "euclidean",
    *,
    threads: int = 1,
    dtype: Optional[Union[_IntegralType, _FloatType, _ComplexType]] = None,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType, _ComplexType] = "d",
) -> Optional[Union[float, complex, DistancesTensor]]: ...

# ---------------------------------------------------------------------
# Vector-vector dot products for real and complex numbers
# ---------------------------------------------------------------------

# Inner product, similar to: `numpy.inner`.
# https://numpy.org/doc/stable/reference/generated/numpy.inner.html
def inner(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[Union[_FloatType, _ComplexType]] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType, _ComplexType] = "d",
) -> Optional[Union[float, complex, DistancesTensor]]: ...

# Dot product, similar to: `numpy.dot`.
# https://numpy.org/doc/stable/reference/generated/numpy.dot.html
def dot(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[Union[_FloatType, _ComplexType]] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType, _ComplexType] = None,
) -> Optional[Union[float, complex, DistancesTensor]]: ...

# Vector-vector dot product for complex conjugates, similar to: `numpy.vdot`.
# https://numpy.org/doc/stable/reference/generated/numpy.vdot.html
def vdot(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[_ComplexType] = None,
    *,
    out: Optional[Union[float, complex, DistancesTensor]] = None,
    out_dtype: Optional[_ComplexType] = None,
) -> Optional[Union[complex, DistancesTensor]]: ...

# ---------------------------------------------------------------------
# Vector-vector spatial distance metrics for real and integer numbers
# ---------------------------------------------------------------------

# Vector-vector squared Euclidean distance, similar to: `scipy.spatial.distance.sqeuclidean`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.sqeuclidean.html
# https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html
def sqeuclidean(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[Union[_IntegralType, _FloatType]] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType] = None,
) -> Optional[Union[float, DistancesTensor]]: ...

# Vector-vector cosine distance, similar to: `scipy.spatial.distance.cosine`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.cosine.html
def cosine(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[Union[_IntegralType, _FloatType]] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType] = None,
) -> Optional[Union[float, DistancesTensor]]: ...

# ---------------------------------------------------------------------
# Vector-vector similarity functions for binary vectors
# ---------------------------------------------------------------------

# Vector-vector Hamming distance, similar to: `scipy.spatial.distance.hamming`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.hamming.html
def hamming(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[_IntegralType] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType] = None,
) -> Optional[Union[float, DistancesTensor]]: ...

# Vector-vector Jaccard distance, similar to: `scipy.spatial.distance.jaccard`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.jaccard.html
def jaccard(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[_IntegralType] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType] = None,
) -> Optional[Union[float, DistancesTensor]]: ...

# ---------------------------------------------------------------------
# Vector-vector similarity between probability distributions
# ---------------------------------------------------------------------

# Vector-vector Jensen-Shannon distance, similar to: `scipy.spatial.distance.jensenshannon`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.jensenshannon.html
def jensenshannon(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[_FloatType] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType] = None,
) -> Optional[Union[float, DistancesTensor]]: ...

# Vector-vector Kullback-Leibler divergence, similar to: `scipy.spatial.distance.kullback_leibler`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.kullback_leibler.html
def kullbackleibler(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[_FloatType] = None,
    *,
    out: Optional[_BufferType] = None,
    out_dtype: Union[_FloatType] = None,
) -> Optional[Union[float, DistancesTensor]]: ...

# ---------------------------------------------------------------------
# Vector-vector similarity between vectors in curved spaces
# ---------------------------------------------------------------------

# Vector-vector bilinear distance, similar to: `numpy.dot(a, metric_tensor @ vector2)`.
# https://numpy.org/doc/stable/reference/generated/numpy.dot.html
def bilinear(
    a: _BufferType,
    b: _BufferType,
    metric_tensor: _BufferType,
    /,
    dtype: Optional[_FloatType] = None,
) -> float: ...

# Vector-vector Mahalanobis distance, similar to: `scipy.spatial.distance.mahalanobis`.
# https://docs.scipy.org/doc/scipy-1.11.4/reference/generated/scipy.spatial.distance.mahalanobis.html
def mahalanobis(
    a: _BufferType,
    b: _BufferType,
    inverse_covariance: _BufferType,
    /,
    dtype: Optional[_FloatType] = None,
) -> float: ...

# ---------------------------------------------------------------------
# Vector-vector similarity between sparse vectors
# ---------------------------------------------------------------------

# Vector-vector intersection similarity, similar to: `numpy.intersect1d`.
# https://numpy.org/doc/stable/reference/generated/numpy.intersect1d.html
def intersection(array1: _BufferType, array2: _BufferType, /) -> float: ...

# ---------------------------------------------------------------------
# Vector-vector math: FMA, WSum
# ---------------------------------------------------------------------

# Vector-vector element-wise fused-multiply add.
def fma(
    a: _BufferType,
    b: _BufferType,
    c: _BufferType,
    /,
    dtype: Optional[Union[_FloatType, _IntegralType]] = None,
    *,
    alpha: float = 1,
    beta: float = 1,
    out: Optional[_BufferType] = None,
) -> Optional[DistancesTensor]: ...

# Vector-vector element-wise weighted sum.
def wum(
    a: _BufferType,
    b: _BufferType,
    /,
    dtype: Optional[Union[_FloatType, _IntegralType]] = None,
    *,
    alpha: float = 1,
    beta: float = 1,
    out: Optional[_BufferType] = None,
) -> Optional[DistancesTensor]: ...



import os
import sys
from os.path import dirname
from setuptools import setup, Extension
from pathlib import Path


__lib_name__ = "simsimd"
__version__ = open("VERSION", "r").read().strip()

compile_args = []
link_args = []
macros_args = [
    ("SIMSIMD_NATIVE_F16", "0"),
    ("SIMSIMD_NATIVE_BF16", "0"),
    ("SIMSIMD_DYNAMIC_DISPATCH", "1"),
]


def is_editable_install():
    """Check if the package is installed in editable mode."""
    if "develop" in sys.argv or "install" in sys.argv and "-e" in sys.argv:
        return True

    # Get the path to the site-packages or dist-packages
    for path in sys.path:
        egg_link = Path(path) / f"{__lib_name__}.egg-link"
        if egg_link.exists():
            return True
    return False


# Check if the installation is in editable mode:
# In that case, we can't include type annotations, as it will lead to name resolution issues.
setup_kwargs = (
    {
        "packages": ["simsimd"],
        "package_dir": {"simsimd": "python/annotations"},
        "package_data": {"simsimd": ["__init__.pyi", "py.typed"]},
    }
    if not is_editable_install()
    else {}
)

if is_editable_install():
    print("You are installing in editable mode. Skipping type annotations installation.")


def get_bool_env(name: str, preference: bool) -> bool:
    return os.environ.get(name, "1" if preference else "0") == "1"


def get_bool_env_w_name(name: str, preference: bool) -> tuple:
    return name, "1" if get_bool_env(name, preference) else "0"


if sys.platform == "linux":
    compile_args.append("-std=c11")
    compile_args.append("-O3")
    compile_args.append("-ffast-math")
    compile_args.append("-fdiagnostics-color=always")
    compile_args.append("-fvisibility=default")
    compile_args.append("-fPIC")
    link_args.append("-shared")

    # Disable warnings
    compile_args.append("-w")

    # Enable OpenMP for Linux
    compile_args.append("-fopenmp")
    link_args.append("-fopenmp")

    # Add vectorized `logf` implementation from the `glibc`
    link_args.append("-lm")

    # SIMD all the way on Linux!
    macros_args.extend(
        [
            get_bool_env_w_name("SIMSIMD_TARGET_NEON", True),
            get_bool_env_w_name("SIMSIMD_TARGET_NEON_F16", True),
            get_bool_env_w_name("SIMSIMD_TARGET_NEON_BF16", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE_F16", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE_BF16", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE2", True),
            get_bool_env_w_name("SIMSIMD_TARGET_HASWELL", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SKYLAKE", True),
            get_bool_env_w_name("SIMSIMD_TARGET_ICE", True),
            get_bool_env_w_name("SIMSIMD_TARGET_GENOA", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SAPPHIRE", True),
            get_bool_env_w_name("SIMSIMD_TARGET_TURIN", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SIERRA", False),  # TODO: Add target spec to GCC & Clang
        ]
    )

if sys.platform == "darwin":
    compile_args.append("-std=c11")
    compile_args.append("-O3")
    compile_args.append("-ffast-math")

    # Disable warnings
    compile_args.append("-w")

    # We can't SIMD all the way on MacOS :(
    macros_args.extend(
        [
            get_bool_env_w_name("SIMSIMD_TARGET_NEON", True),
            get_bool_env_w_name("SIMSIMD_TARGET_NEON_F16", True),  # Supported on Apple M1 and newer
            get_bool_env_w_name("SIMSIMD_TARGET_NEON_BF16", True),  # Supported on Apple M2 and newer
            get_bool_env_w_name("SIMSIMD_TARGET_SVE", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE2", False),
            get_bool_env_w_name("SIMSIMD_TARGET_HASWELL", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SKYLAKE", False),
            get_bool_env_w_name("SIMSIMD_TARGET_ICE", False),
            get_bool_env_w_name("SIMSIMD_TARGET_GENOA", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SAPPHIRE", False),
            get_bool_env_w_name("SIMSIMD_TARGET_TURIN", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SIERRA", False),
        ]
    )

if sys.platform == "win32":
    compile_args.append("/std:c11")
    compile_args.append("/O2")
    compile_args.append("/fp:fast")
    compile_args.append("/EXPORT:*")

    # Dealing with MinGW linking errors
    # https://cibuildwheel.readthedocs.io/en/stable/faq/#windows-importerror-dll-load-failed-the-specific-module-could-not-be-found
    compile_args.append("/d2FH4-")

    # We can't SIMD all the way on Windows :(
    # Even NEON `f16` fails: https://github.com/ashvardanian/SimSIMD/actions/runs/11419164624/job/31773473319?pr=214
    macros_args.extend(
        [
            get_bool_env_w_name("SIMSIMD_TARGET_NEON", True),
            get_bool_env_w_name("SIMSIMD_TARGET_NEON_F16", False),
            get_bool_env_w_name("SIMSIMD_TARGET_NEON_BF16", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SVE2", False),
            get_bool_env_w_name("SIMSIMD_TARGET_HASWELL", True),
            get_bool_env_w_name("SIMSIMD_TARGET_SKYLAKE", True),
            get_bool_env_w_name("SIMSIMD_TARGET_ICE", True),
            get_bool_env_w_name("SIMSIMD_TARGET_GENOA", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SAPPHIRE", False),
            get_bool_env_w_name("SIMSIMD_TARGET_TURIN", False),
            get_bool_env_w_name("SIMSIMD_TARGET_SIERRA", False),
        ]
    )

ext_modules = [
    Extension(
        "simsimd",
        sources=["python/lib.c", "c/lib.c"],
        include_dirs=["include"],
        extra_compile_args=compile_args,
        extra_link_args=link_args,
        define_macros=macros_args,
    ),
]

this_directory = os.path.abspath(dirname(__file__))
with open(os.path.join(this_directory, "README.md"), "r", encoding="utf8") as f:
    long_description = f.read()

setup(
    name=__lib_name__,
    version=__version__,
    author="Ash Vardanian",
    author_email="1983160+ashvardanian@users.noreply.github.com",
    url="https://github.com/ashvardanian/simsimd",
    description="Portable mixed-precision BLAS-like vector math library for x86 and ARM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Programming Language :: C",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    ext_modules=ext_modules,
    zip_safe=False,
    include_package_data=True,
    **setup_kwargs,
)
